from octopipes.process_debuging import echo_bbox_outofbound


def test_echo_bbox_outofbound(capfd):
    echo_bbox_outofbound(100, 100)([(0, 0, 0, 0)])
    out, _ = capfd.readouterr()
    assert out == ''

    echo_bbox_outofbound(50, 50)([(50, 50, 100, 100)])
    out, _ = capfd.readouterr()
    assert out.startswith('Out of bound bbox')



import logging
import json
from typing import Any, Protocol

import cmap


logger = logging.getLogger(__name__)

class HandlerInterface(Protocol):
    def output_on_image(self, image, output):
        pass

    def len_output(self, output: Any) -> int | None:
        pass

    def to_json(self, output: Any) -> str: # type: ignore
        pass


class DefaultHandler:
    def output_on_image(self, image, output):
        return image

    def len_output(self, output: Any) -> int | None:
        return len(output)

    def to_json(self, output: Any) -> str:
        return json.dumps(output)


class SegmentationMasksHandler:
    def output_on_image(self, image, output):
        import numpy as np
        import cv2

        sorted_anns = sorted(output, key=(lambda x: x['area']), reverse=True)
        w, h = sorted_anns[0]['segmentation'].shape[0], sorted_anns[0]['segmentation'].shape[1]
        overlay_mask = np.zeros((w, h, 3), dtype=np.uint8)
        for ann in sorted_anns:
            mask = ann['segmentation']
            overlay_mask[mask] = np.random.randint(256, size=3, dtype=np.uint8)
        overlayed = cv2.addWeighted(image, 1, overlay_mask, 0.3, 20)
        return overlayed

    def to_json(self, output) -> str:
        return json.dumps(output.tolist())

    def len_output(self, output: Any) -> int | None:
        return len(output)


class BboxesHandler:
    def output_on_image(self, image, output):
        import cv2
        import numpy as np

        if output is None:
            return image
        color = list(np.random.choice(range(256), size=3))
        for bbox in output:
            cv2.rectangle(image, (bbox[0], bbox[1]), (bbox[2], bbox[3]), color, 3)

        return image

    def to_json(self, output) -> str:
        try:
            # convert to list if object is np array
            output = output.tolist()
        except AttributeError:
            pass
        return json.dumps({'bboxes': [{'bbox': bbox} for bbox in output] if output is not None else None})

    def len_output(self, output: Any) -> int | None:
        return len(output)


class CmapBboxesHandler:
    """CmapBboxesHandler colors bboxes with a colormap.
    Output is expected to be [(value, bbox), ...]"""

    def output_on_image(self, image, output):
        import cv2
        import numpy as np

        values, bboxes = zip(*output)
        values = np.array(values)
        norm_values = (values - values.min()) / (values.max() - values.min())
        cm = cmap.Colormap('viridis')(norm_values)

        for i, bbox in enumerate(bboxes):
            color = np.array(255 * cm[i, :3], dtype=int)
            cv2.rectangle(image,
                          (bbox[0], bbox[1]), (bbox[2], bbox[3]),
                          (color[0], color[1], color[2]), 3)
        
        return image

    def to_json(self, output) -> str:
        try:
            # convert to list if object is np array
            output = output.tolist()
        except AttributeError:
            pass
        return json.dumps({'bboxes': [{'bbox': bbox, 'val': val} for bbox, val in output] if output is not None else None})

    def len_output(self, output: Any) -> int | None:
        return len(output)


class CirclesHandler:
    """CirclesHandler handles output of the form [(x, y, radius), ...]"""

    def output_on_image(self, image, output):
        import cv2
        import numpy as np

        if output is None:
            return image

        for x, y, r in output:
            color = list(np.random.choice(range(256), size=3))
            cv2.circle(image, (int(x), int(y)), int(r), color, 3)
        return image

    def to_json(self, output) -> str:
        try:
            # convert to list if object is np array
            output = output.tolist()
        except AttributeError:
            pass
        return json.dumps({'circles': output})

    def len_output(self, output: Any) -> int | None:
        return len(output)




from octopipes.benchmark import Benchmark
from octopipes.dataset import Dataloader, Dataset
from octopipes.workflow import Workflow

from tests.test_dataset import MockDataset


def test_benchmark():
    wf1 = Workflow('test_wf_1')\
            .add(lambda x: x + 1)
    wf2 = Workflow('test_wf_2')\
            .add(lambda x: x * 2)
    dataset: Dataset = MockDataset([1, 2, 3, 4, 5])
    dataloader = Dataloader(dataset=dataset, batch_size=2, drop_last_batch=True)
    benchmark = Benchmark(dataloader=dataloader, workflows=[wf1, wf2])
    benchmark.run_tests()
    assert len(benchmark.results) == 4
    assert benchmark.results[0].results[0].output == 2
    assert benchmark.results[0].results[1].output == 2
    assert benchmark.results[1].results[0].output == 3
    assert benchmark.results[1].results[1].output == 4



import logging
from collections.abc import Callable
from typing import Protocol

from tqdm import tqdm

from octopipes.results import Results
from octopipes.workflow import Workflow

logger = logging.getLogger(__name__)

class AggregateFlows:
    """AggregateFlows enables running multiple workflows on the same input"""

    def __init__(self, input, workflows: list[Workflow]):
        self.input = input
        self.workflows = workflows
        self.results: list[Results] = []
        self._hooks = []

    def add_hook(self, hook: Callable):
        """Adds a hook for post-workflow callbacks"""
        self._hooks.append(hook)

    def add_hooks(self, hooks: list[Callable]):
        """Adds multiple hooks for post-workflow callbacks"""
        [self.add_hook(h) for h in hooks]

    def _run_hook(self, hook, workflow):
        hook(workflow)

    def run_hooks(self, workflow):
        for hook in self._hooks:
            self._run_hook(hook, workflow)

    def _run(self, workflow: Workflow) -> Results:
        wf_iter = workflow(self.input)
        for _ in tqdm(wf_iter, desc=f'{workflow.name} Steps', leave=False):
            pass

        result = wf_iter.freeze()
        self.results.append(result)

        # running hooks
        self.run_hooks(wf_iter)
        return result

    def run_workflows(self):
        for wf in tqdm(self.workflows, desc='workflows', leave=False):
            self._run(wf)


class AggregateFlowsFactory(Protocol):
    def get_aggregate_flows(self, input, workflows) -> AggregateFlows: # type: ignore
        pass

class DefaultAggregateFlowsFactory:
    def __init__(self, hooks: list[Callable]):
        self.hooks = hooks

    def get_aggregate_flows(self, input, workflows) -> AggregateFlows:
        aggr = AggregateFlows(input, workflows=workflows)
        aggr.add_hooks(self.hooks)
        return aggr



from octopipes.process import circles_to_bboxes


def test_circles_to_bboxes():
    assert circles_to_bboxes([]) == []
    assert circles_to_bboxes([(0, 0, 0)]) == [(0, 0, 0, 0)]
    assert circles_to_bboxes([(0, 0, 0), (0, 0, 0)]) == [(0, 0, 0, 0), (0, 0, 0, 0)]



import numpy as np

from octopipes.handlers import BboxesHandler, CirclesHandler, CmapBboxesHandler, DefaultHandler


def test_DefaultHandler():
    handler = DefaultHandler()
    assert handler.output_on_image(None, None) is None
    assert handler.len_output([1, 2, 3]) == 3
    assert handler.to_json([1, 2, 3]) == '[1, 2, 3]'


def test_BboxHandler():
    handler = BboxesHandler()
    test_list = [[0, 0, 10, 10], [0, 0, 10, 10]]
    test_array = np.array(test_list)

    assert handler.len_output([[0, 0, 10, 10]]) == 1
    assert handler.len_output(test_list) == 2
    assert handler.len_output(test_array) == 2

    assert handler.to_json([[0, 0, 10, 10]]) == '{"bboxes": [{"bbox": [0, 0, 10, 10]}]}'
    assert handler.to_json(np.array([[0, 0, 10, 10]])) == '{"bboxes": [{"bbox": [0, 0, 10, 10]}]}'
    assert handler.to_json([[]]) == '{"bboxes": [{"bbox": []}]}'
    assert handler.to_json(np.array([[]])) == '{"bboxes": [{"bbox": []}]}'
    assert handler.to_json(np.array([])) == '{"bboxes": []}'
    assert handler.to_json(None) == '{"bboxes": null}'


def test_CirclesHandler():
    handler = CirclesHandler()
    test_list = [[0, 0, 10], [10, 7, 5]]
    test_array = np.array(test_list)

    assert handler.len_output(test_list) == 2
    assert handler.len_output(test_array) == 2

    assert handler.to_json(test_list) == '{"circles": [[0, 0, 10], [10, 7, 5]]}'
    assert handler.to_json(test_array) == '{"circles": [[0, 0, 10], [10, 7, 5]]}'
    assert handler.to_json([]) == '{"circles": []}'
    assert handler.to_json(None) == '{"circles": null}'


def test_CmapBboxesHandler():
    handler = CmapBboxesHandler()
    test_list = [([0, 0, 10, 10], 1), ([0, 0, 10, 10], 2)]

    assert handler.len_output(test_list) == 2

    assert handler.to_json(test_list) == '{"bboxes": [{"bbox": [0, 0, 10, 10], "val": 1}, {"bbox": [0, 0, 10, 10], "val": 2}]}'
    assert handler.to_json([]) == '{"bboxes": []}'
    assert handler.to_json(None) == '{"bboxes": null}'




from octopipes.workflow import Workflow


def test_workflow():
    wf = Workflow('test_wf_1')\
            .add(lambda x: x)\
            .add(lambda x: x)
    assert wf.nsteps == 2
    wf_iter = wf(1)
    for result in wf_iter:
        assert result == 1

    wf_iter.recap()
    wf_iter.freeze()



from dataclasses import dataclass
from typing import Protocol, Any


class Dataset(Protocol):
    def __getitem__(self, index) -> Any:
        pass

    def __len__(self) -> int:
        return 0


@dataclass
class Dataloader:
    dataset: Dataset
    batch_size: int = 1
    limit: int | None = None
    drop_last_batch: bool = False

    def __iter__(self):
        return iter(DataloaderIter(self))


class DataloaderIter:
    def __init__(self, dataloader) -> None:
        self.dataset = dataloader.dataset
        self.batch_size = dataloader.batch_size
        self.limit = dataloader.limit
        self.drop_last_batch = dataloader.drop_last_batch

        self.size = len(self.dataset)
        self.batches = self.size // self.batch_size
        self.total_yield = self.batch_size * self.batches
        if not dataloader.drop_last_batch:
            self.batches += 1 if self.size % self.batch_size != 0 else 0
            self.total_yield += self.size % self.batch_size
        if self.limit is not None:
            self.total_yield = self.limit

    def __iter__(self):
        self.current_index = 0
        self.total_yielded = 0
        return self

    def __next__(self):
        if self.total_yielded < self.total_yield:
            end = self.current_index + self.batch_size
            batch = self.dataset[self.current_index:end]
            if self.total_yielded + len(batch) > self.total_yield:
                batch = batch[:self.total_yield - self.total_yielded]
            self.current_index = end
            self.total_yielded += len(batch)
            return batch
        raise StopIteration





from octopipes.aggregate_flows import AggregateFlows
from octopipes.workflow import Workflow


def test_aggregateflows():
    wf1 = Workflow('test_wf_1')\
            .add(lambda x: x + 1)
    wf2 = Workflow('test_wf_2')\
            .add(lambda x: x * 2)
    flows = AggregateFlows(2, workflows=[wf1, wf2])
    flows.run_workflows()
    assert len(flows.results) == 2
    assert flows.results[0].output == 3
    assert flows.results[1].output == 4




from typing import Any
from typing_extensions import TypedDict

from pydantic import BaseModel, ConfigDict


class Step(TypedDict):
    step: str
    duration: float


class Results(BaseModel):
    model_config = ConfigDict(frozen=True)

    name: str
    metadata_name: str
    metadata: dict
    nsteps: int
    current_step: int
    output: Any
    len_output: int | None
    total_duration: float
    output_recap: tuple[Step, ...]



"""General debuging processes for workflow troubleshooting"""

def echo_output(input):
    print(input)
    return input

def echo_bbox_outofbound(width: int, height: int):
    """Echo out of bound bboxes"""
    def check_bbox_outofbound(bboxes):
        for x, y, maxx, maxy in bboxes:
            if (x > width or maxx > width or 
                    y > height or maxy > height):
                print(f'Out of bound bbox {(x, y, maxx, maxy)}')
        return bboxes
    return check_bbox_outofbound



from multiprocess import Pool

from tqdm import tqdm

from octopipes.dataset import Dataloader
from octopipes.workflow import Workflow
from octopipes.aggregate_flows import AggregateFlows, AggregateFlowsFactory, DefaultAggregateFlowsFactory


class Benchmark:
    def __init__(self, dataloader: Dataloader, workflows: list[Workflow],
                 flows_factory: AggregateFlowsFactory | None = None) -> None:
        self.dataloader = dataloader
        self.workflows = workflows
        self.results: list[AggregateFlows] = []

        self.factory: AggregateFlowsFactory = DefaultAggregateFlowsFactory(hooks=[]) if flows_factory is None else flows_factory

    @staticmethod
    def run_sample(factory: AggregateFlowsFactory, workflows, sample):
        try:
            feature, _ = sample
        except TypeError:
            feature = sample

        aggregate = factory.get_aggregate_flows(feature, workflows)
        aggregate.run_workflows()
        return aggregate

    def run_tests(self):
        for batch in tqdm(self.dataloader):
            with Pool(processes=len(batch)) as pool:
                for result in pool.map(func=Run(self.factory, self.workflows), iterable=batch):
                    self.results.append(result)


class Run:
    def __init__(self, factory, workflows) -> None:
        self.factory = factory
        self.workflows = workflows

    def __call__(self, sample):
        return Benchmark.run_sample(self.factory, self.workflows, sample)



import pytest

from octopipes.dataset import Dataloader, Dataset

class MockDataset:
    def __init__(self, values):
        self.l = values

    def __len__(self):
        return len(self.l)

    def __getitem__(self, index):
        return self.l[index]


def test_dataset():
    dataset: Dataset = MockDataset([])
    assert len(dataset) == 0

    dataset: Dataset = MockDataset([1, 2, 3, 4, 5])
    assert len(dataset) == 5
    assert dataset[0] == 1
    assert dataset[4] == 5

    dataset: Dataset = MockDataset([([1, 2, 3], 1)])
    assert len(dataset) == 1
    assert dataset[0] == ([1, 2, 3], 1)

    dataset: Dataset = MockDataset([([1, 2, 3], 1), ([4, 5, 6], 2)])
    assert dataset[1][1] == 2


def test_dataloader():
    dataset: Dataset = MockDataset([])
    dataloader = Dataloader(dataset=dataset,
                            batch_size=1,
                            limit=1)
    batch_iter = iter(dataloader)
    assert next(batch_iter) == []

    dataset: Dataset = MockDataset([1, 2, 3, 4, 5])
    dataloader = Dataloader(dataset=dataset,
                            batch_size=2)
    batch_iter = iter(dataloader)
    assert next(batch_iter) == [1, 2]
    assert next(batch_iter) == [3, 4]
    assert next(batch_iter) == [5]
    with pytest.raises(StopIteration):
        next(batch_iter)

    # Test when limit = batch_size
    dataloader = Dataloader(dataset=dataset,
                            batch_size=3,
                            limit=3)
    batch_iter = iter(dataloader)
    assert next(batch_iter) == [1, 2, 3]
    with pytest.raises(StopIteration):
        next(batch_iter)

    # Test when batch_size > limit
    dataloader = Dataloader(dataset=dataset,
                            batch_size=3,
                            limit=2)
    batch_iter = iter(dataloader)
    assert next(batch_iter) == [1, 2]
    with pytest.raises(StopIteration):
        next(batch_iter)

    # Test when limit > batch_size
    dataloader = Dataloader(dataset=dataset,
                            batch_size=3,
                            limit=4)
    batch_iter = iter(dataloader)
    assert next(batch_iter) == [1, 2, 3]
    assert next(batch_iter) == [4]
    with pytest.raises(StopIteration):
        next(batch_iter)



import time
import logging
from typing import Any
from collections.abc import Callable

from octopipes.results import Results, Step
from octopipes.handlers import DefaultHandler, HandlerInterface


logger = logging.getLogger(__name__)


class Workflow:
    """Workflow allows the difinition of multiple processes/steps"""
    
    def __init__(self, name: str,
                 steps: list[tuple[Callable, HandlerInterface]] = list(),
                 metadata: dict = {}) -> None:
        self.name = name
        self.metadata = dict(metadata)
        
        try:
            processes, handlers = zip(*list(steps))
        except ValueError:
            processes, handlers = [], []

        self.processes: list[Callable] = list(processes)  # type: ignore
        self.handlers: list[HandlerInterface] = list(handlers)  # type: ignore

    @property
    def metadata_name(self):
        return f'{self.name}{"".join(f"({key}:{value})" for key, value in self.metadata)}'

    @property
    def nsteps(self):
        return len(self.processes)

    def __len__(self):
        return self.nsteps

    def add(self, process: Callable, handler: HandlerInterface = DefaultHandler()):
        self.processes.append(process)
        self.handlers.append(handler)
        return self

    def __call__(self, input: Any, *args: Any, **kwds: Any) -> 'WorkflowIter':
        return Workflow.WorkflowIter(self, input)

    class WorkflowIter:
        def __init__(self, workflow: 'Workflow', input) -> None:
            self.workflow = workflow
            self.input = input
            self.outputs = []
            self.durations : list[float] = []

        def __iter__(self):
            self.current_step = 0
            self.current_output = self.input
            return self

        def __next__(self):
            if self.current_step < self.workflow.nsteps:
                self.current_step += 1

                start = time.perf_counter()
                self.current_output = self.workflow.processes[self.current_step - 1](self.current_output)
                end = time.perf_counter()
                self.outputs.append(self.current_output)
                self.durations.append(end - start)
                return self.current_output
            raise StopIteration

        def recap(self, capture=False):
            output = ''
            for process, d in zip(self.workflow.processes, self.durations):
                output += f'{process.__name__} took {d:.4f}s'
            if capture:
                return output
            print(output)

        def steps_recap(self) -> tuple[Step, ...]:
            return tuple({'step': process.__name__, 'duration': d} for process, d in zip(self.workflow.processes, self.durations))

        @property
        def total_duration(self):
            return sum(self.durations)

        @property
        def final_output(self):
            return self.outputs[-1]

        def len_output(self):
            return len(self.current_output) if hasattr(self.current_output, '__len__') else None

        def freeze(self) -> Results:
            return Results(name=self.workflow.name,
                           metadata_name=self.workflow.metadata_name,
                           metadata=self.workflow.metadata,
                           nsteps=self.workflow.nsteps,
                           current_step=self.current_step,
                           output=self.current_output,
                           len_output=self.len_output(),
                           total_duration=self.total_duration,
                           output_recap=self.steps_recap())



import pathlib
from setuptools import find_packages
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    name='octopipes',
    version='0.1.0-1',
    description='Pipeline library for AI workflows.',
    author='Octomiro',
    author_email='contact@octomiro.ai',
    url='https://github.com/octomiro/octopipes',
    long_description=README,
    long_description_content_type='text/markdown',
    packages=find_packages(exclude=('tests*', 'testing*')),
    install_requires=[
        'pydantic',
        'multiprocess',
        'numpy',
        'tqdm',
        'cmap'
    ],
    extras_require={
        'opencv': ['opencv-python'],
        'pytorch': ['torch', 'torchvision']
    },
    classifiers=[
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.10',
        'License :: OSI Approved :: MIT License'
    ]
)
